@@ -26,6 +26,7 @@
 
 
 typedef struct {
+    struct bt_ref_hdr hdr;
     CK_SLOT_ID slotID;
     CK_SESSION_HANDLE sessionh;
 } ST_SESSION_T;
@@ -87,7 +88,8 @@ typedef CK_RV (CK_PTR ST_C_OpenSession) (STDLL_TokData_t *tokdata,
                                          CK_SESSION_HANDLE_PTR phSession);
 
 typedef CK_RV (CK_PTR ST_C_CloseSession) (STDLL_TokData_t *tokdata,
-                                          ST_SESSION_T *hSession);
+                                          ST_SESSION_T *hSession,
+                                          CK_BBOOL in_fork_initializer);
 typedef CK_RV (CK_PTR ST_C_CloseAllSessions) (STDLL_TokData_t *tokdata,
                                               CK_SLOT_ID slotID);
 typedef CK_RV (CK_PTR ST_C_GetSessionInfo) (STDLL_TokData_t *tokdata,
@@ -337,6 +339,16 @@ typedef CK_RV (CK_PTR ST_C_WaitForSlotEvent) (STDLL_TokData_t *tokdata,
                                               CK_SLOT_ID_PTR pSlot,
                                               CK_VOID_PTR pReserved);
 
+typedef CK_RV (CK_PTR ST_C_IBM_ReencryptSingle)(STDLL_TokData_t *tokdata,
+                                                ST_SESSION_T *hSession,
+                                                CK_MECHANISM_PTR pDecrMech,
+                                                CK_OBJECT_HANDLE hDecrKey,
+                                                CK_MECHANISM_PTR pEncrMech,
+                                                CK_OBJECT_HANDLE hEncrKey,
+                                                CK_BYTE_PTR pEncryptedData,
+                                                CK_ULONG ulEncryptedDataLen,
+                                                CK_BYTE_PTR pReencryptedData,
+                                            CK_ULONG_PTR pulReencryptedDataLen);
 
 
 struct ST_FCN_LIST {
@@ -411,6 +423,7 @@ struct ST_FCN_LIST {
     ST_C_GetFunctionStatus ST_GetFunctionStatus;
     ST_C_CancelFunction ST_CancelFunction;
 
+    ST_C_IBM_ReencryptSingle ST_IBM_ReencryptSingle;
 };
 
 typedef struct ST_FCN_LIST STDLL_FcnList_t;
