@@ -32,7 +32,7 @@
 token_spec_t token_specific = {
     CCA_CONFIG_PATH,
     "ccatok",
-    64,
+    TRUE,
     // Token data info:
     {
         FALSE,                     // Don't use per guest data store
@@ -76,10 +76,10 @@ token_spec_t token_specific = {
     NULL,                       // rsa_x509_sign
     NULL,                       // rsa_x509_verify
     NULL,                       // rsa_x509_verify_recover
-    NULL,                       // rsa_oaep_decrypt
-    NULL,                       // rsa_oaep_encrypt
-    NULL,                       // rsa_pss_sign
-    NULL,                       // rsa_pss_verify
+    &token_specific_rsa_oaep_decrypt,
+    &token_specific_rsa_oaep_encrypt,
+    &token_specific_rsa_pss_sign,
+    &token_specific_rsa_pss_verify,
     &token_specific_rsa_generate_keypair,
     // Elliptic Curve
     &token_specific_ec_sign,
@@ -128,7 +128,10 @@ token_spec_t token_specific = {
     NULL,                       // dsa_verify
     &token_specific_get_mechanism_list,
     &token_specific_get_mechanism_info,
-    &token_specific_object_add
+    &token_specific_object_add,
+    &token_specific_key_wrap,
+    &token_specific_key_unwrap,
+    &token_specific_reencrypt_single,
 };
 
 #endif
