@@ -41,6 +41,7 @@
 #include "sw_crypt.h"
 #include "trace.h"
 #include "ock_syslog.h"
+#include "slotmgr.h" // for ock_snprintf
 
 extern void set_perm(int);
 
@@ -50,18 +51,71 @@ CK_RV reload_token_object_old(STDLL_TokData_t *tokdata, OBJECT *obj);
 CK_RV save_public_token_object_old(STDLL_TokData_t *tokdata, OBJECT *obj);
 CK_RV load_public_token_objects_old(STDLL_TokData_t *tokdata);
 
-char *pk_dir;
-
-char *get_pk_dir(char *fname)
+static int get_token_object_path(char *buf, size_t buflen,
+                                 STDLL_TokData_t *tokdata, char *path)
 {
+    if (ock_snprintf(buf, buflen, "%s/" PK_LITE_OBJ_DIR "/%s",
+                     tokdata->data_store, path) != 0) {
+        TRACE_ERROR("buffer overflow for object path %s", path);
+        return -1;
+    }
+    return 0;
+}
+
+static FILE *open_token_object_path(char *buf, size_t buflen,
+                                    STDLL_TokData_t *tokdata, char *path,
+                                    char *mode)
+{
+    if (get_token_object_path(buf, buflen, tokdata, path) < 0)
+        return NULL;
+    return fopen(buf, mode);
+}
+
+static int get_token_data_store_path(char *buf, size_t buflen,
+                                     STDLL_TokData_t *tokdata, char *path)
+{
+    if (ock_snprintf(buf, buflen, "%s/%s", tokdata->data_store, path)) {
+        TRACE_ERROR("buffer overflow for path %s", path);
+        return -1;
+    }
+    return 0;
+}
+
+static FILE *open_token_data_store_path(char *buf, size_t buflen,
+                                        STDLL_TokData_t *tokdata, char *path,
+                                        char *mode)
+{
+    if (get_token_data_store_path(buf, buflen, tokdata, path) < 0)
+        return NULL;
+    return fopen(buf, mode);
+}
+
+static FILE *open_token_object_index(char *buf, size_t buflen,
+                                     STDLL_TokData_t *tokdata, char *mode)
+{
+    return open_token_object_path(buf, buflen, tokdata, PK_LITE_OBJ_IDX, mode);
+}
+
+static FILE *open_token_nvdat(char *buf, size_t buflen,
+                              STDLL_TokData_t *tokdata, char *mode)
+{
+    if (ock_snprintf(buf, buflen, "%s/" PK_LITE_NV, tokdata->data_store)) {
+        TRACE_ERROR("NVDAT.TOK file name buffer overflow\n");
+        return NULL;
+    }
+    return fopen(buf, mode);
+}
+
+char *get_pk_dir(STDLL_TokData_t *tokdata, char *fname, size_t len)
+{
+    int snres;
     struct passwd *pw = NULL;
 
     if (token_specific.data_store.per_user && (pw = getpwuid(getuid())) != NULL)
-        sprintf(fname, "%s/%s", pk_dir, pw->pw_name);
+        snres = ock_snprintf(fname, len, "%s/%s", tokdata->pk_dir, pw->pw_name);
     else
-        sprintf(fname, "%s", pk_dir);
-
-    return fname;
+        snres = ock_snprintf(fname, len, "%s", tokdata->pk_dir);
+    return snres != 0 ? NULL : fname;
 }
 
 void set_perm(int file)
@@ -96,6 +150,7 @@ error:
 
 //
 // Note: The token lock (XProcLock) must be held when calling this function.
+// The object must hold the READ lock when this function is called.
 //
 CK_RV save_token_object(STDLL_TokData_t *tokdata, OBJECT *obj)
 {
@@ -113,9 +168,7 @@ CK_RV save_token_object(STDLL_TokData_t *tokdata, OBJECT *obj)
         return rc;
 
     // update the index file if it exists
-    sprintf(fname, "%s/%s/%s", tokdata->data_store, PK_LITE_OBJ_DIR,
-            PK_LITE_OBJ_IDX);
-    fp = fopen(fname, "r");
+    fp = open_token_object_index(fname, sizeof(fname), tokdata, "r");
     if (fp) {
         set_perm(fileno(fp));
         while (fgets(line, 50, fp)) {
@@ -152,11 +205,6 @@ CK_RV delete_token_object(STDLL_TokData_t *tokdata, OBJECT *obj)
     FILE *fp1, *fp2;
     char objidx[PATH_MAX], idxtmp[PATH_MAX], fname[PATH_MAX], line[256];
 
-    sprintf(objidx, "%s/%s/%s", tokdata->data_store,
-            PK_LITE_OBJ_DIR, PK_LITE_OBJ_IDX);
-    sprintf(idxtmp, "%s/%s/%s", tokdata->data_store,
-            PK_LITE_OBJ_DIR, "IDX.TMP");
-
     // FIXME:  on UNIX, we need to make sure these guys aren't symlinks
     //         before we blindly write to these files...
     //
@@ -164,8 +212,9 @@ CK_RV delete_token_object(STDLL_TokData_t *tokdata, OBJECT *obj)
     // remove the object from the index file
     //
 
-    fp1 = fopen(objidx, "r");
-    fp2 = fopen(idxtmp, "w");
+    fp1 = open_token_object_index(objidx, sizeof(objidx), tokdata, "r");
+    fp2 = open_token_object_path(idxtmp, sizeof(idxtmp),
+                                 tokdata, "IDX.TMP", "w");
     if (!fp1 || !fp2) {
         if (fp1)
             fclose(fp1);
@@ -207,9 +256,11 @@ CK_RV delete_token_object(STDLL_TokData_t *tokdata, OBJECT *obj)
     fclose(fp1);
     fclose(fp2);
 
-    sprintf(fname, "%s/%s/%s", tokdata->data_store,
-            PK_LITE_OBJ_DIR, (char *) obj->name);
-    unlink(fname);
+    if (get_token_object_path(fname, sizeof(fname), tokdata,
+                              (char *) obj->name) < 0)
+       TRACE_DEVEL("file name buffer overflow in obj unlink\n");
+    else
+        unlink(fname);
 
     return CKR_OK;
 }
@@ -240,42 +291,50 @@ done:
     return rc;
 }
 
-void init_data_store(char *directory, char *data_store)
+CK_RV init_data_store(STDLL_TokData_t *tokdata, char *directory,
+                      char *data_store, size_t len)
 {
     char *pkdir;
+    int pklen;
 
-    if (pk_dir != NULL) {
-        free(pk_dir);
-        pk_dir = NULL;
+    if (tokdata->pk_dir != NULL) {
+        free(tokdata->pk_dir);
+        tokdata->pk_dir = NULL;
     }
 
-    if ((pkdir = getenv("PKCS_APP_STORE")) != NULL) {
-        pk_dir = (char *) malloc(strlen(pkdir) + 1024);
-        memset(pk_dir, 0, strlen(pkdir) + 1024);
-        sprintf(pk_dir, "%s/%s", pkdir, SUB_DIR);
-        get_pk_dir(data_store);
-        return;
+    if ((pkdir = secure_getenv("PKCS_APP_STORE")) != NULL) {
+        pklen = strlen(pkdir) + 1024;
+        tokdata->pk_dir = (char *) calloc(pklen, 1);
+        if (!(tokdata->pk_dir))
+            return CKR_HOST_MEMORY;
+        if (ock_snprintf(tokdata->pk_dir, pklen, "%s/%s", pkdir, SUB_DIR) != 0)
+            return CKR_FUNCTION_FAILED;
+        return get_pk_dir(tokdata, data_store, len) ? CKR_OK : CKR_FUNCTION_FAILED;
     }
 
     if (directory) {
-        pk_dir = (char *) malloc(strlen(directory) + 25);
-        memset(pk_dir, 0, strlen(directory) + 25);
-        sprintf(pk_dir, "%s", directory);
+        pklen = strlen(directory) + 1;
+        tokdata->pk_dir = (char *) calloc(pklen, 1);
+        if (!(tokdata->pk_dir))
+            return CKR_HOST_MEMORY;
+        if (ock_snprintf(tokdata->pk_dir, pklen, "%s", directory) != 0)
+            return CKR_FUNCTION_FAILED;
     } else {
-        pk_dir = (char *) malloc(strlen(PK_DIR) + 25);
-        memset(pk_dir, 0, strlen(PK_DIR) + 25);
-        sprintf(pk_dir, "%s", PK_DIR);
+        pklen = strlen(PK_DIR) + 1;
+        tokdata->pk_dir = (char *) calloc(pklen, 1);
+        if (!(tokdata->pk_dir))
+            return CKR_HOST_MEMORY;
+        if (ock_snprintf(tokdata->pk_dir, pklen, "%s", PK_DIR) != 0)
+            return CKR_FUNCTION_FAILED;
     }
-    get_pk_dir(data_store);
-
-    return;
+    return get_pk_dir(tokdata, data_store, len) ? CKR_OK : CKR_FUNCTION_FAILED;
 }
 
-void final_data_store(void)
+void final_data_store(STDLL_TokData_t * tokdata)
 {
-    if (pk_dir != NULL) {
-        free(pk_dir);
-        pk_dir = NULL;
+    if (tokdata->pk_dir != NULL) {
+        free(tokdata->pk_dir);
+        tokdata->pk_dir = NULL;
     }
 }
 
@@ -283,8 +342,7 @@ void final_data_store(void)
  * tokversion < 3.12 object store
  */
 
-static CK_RV get_encryption_info_for_clear_key(CK_ULONG *p_key_len,
-                                               CK_ULONG *p_block_size)
+static CK_RV get_encryption_info(CK_ULONG *p_key_len, CK_ULONG *p_block_size)
 {
     CK_ULONG key_len = 0L;
     CK_ULONG block_size = 0L;
@@ -311,25 +369,6 @@ static CK_RV get_encryption_info_for_clear_key(CK_ULONG *p_key_len,
     return CKR_OK;
 }
 
-static CK_RV get_encryption_info(CK_ULONG *p_key_len, CK_ULONG *p_block_size)
-{
-    CK_RV rc;
-
-    rc = get_encryption_info_for_clear_key(p_key_len, p_block_size);
-    if (rc != CKR_OK)
-        return rc;
-
-    /* Tokens that use a secure key have a different size for key because
-     * it's just an indentifier not a real key.
-     */
-    if (is_secure_key_token()) {
-        if (p_key_len)
-            *p_key_len = token_specific.token_keysize;
-    }
-
-    return CKR_OK;
-}
-
 static CK_BYTE *duplicate_initial_vector(const CK_BYTE *iv)
 {
     CK_ULONG block_size = 0L;
@@ -430,7 +469,7 @@ static CK_RV encrypt_data_with_clear_key(STDLL_TokData_t *tokdata,
     CK_RV rc = CKR_OK;
     CK_BYTE *initial_vector = NULL;
 
-    if (!is_secure_key_token() &&
+    if (!token_specific.secure_key_token &&
         token_specific.data_store.encryption_algorithm != CKM_DES3_CBC) {
         return encrypt_data(tokdata, key, keylen, iv, clear, clear_len,
                             cipher, p_cipher_len);
@@ -546,7 +585,7 @@ static CK_RV decrypt_data_with_clear_key(STDLL_TokData_t *tokdata,
     CK_RV rc = CKR_OK;
     CK_BYTE *initial_vector = NULL;
 
-    if (!is_secure_key_token() &&
+    if (!token_specific.secure_key_token &&
         token_specific.data_store.encryption_algorithm != CKM_DES3_CBC) {
         return decrypt_data(tokdata, key, keylen, iv, cipher,
                             cipher_len, clear, p_clear_len);
@@ -600,28 +639,12 @@ CK_RV load_token_data_old(STDLL_TokData_t *tokdata, CK_SLOT_ID slot_id)
         goto out_nolock;
     }
 
-    sprintf(fname, "%s/%s", tokdata->data_store, PK_LITE_NV);
-    fp = fopen(fname, "r");
+    fp = open_token_nvdat(fname, sizeof(fname), tokdata, "r");
     if (!fp) {
         /* Better error checking added */
         if (errno == ENOENT) {
-            /* init_token_data may call save_token_data, which
-             * grabs the lock, so we must release it around this
-             * call */
-            rc = XProcUnLock(tokdata);
-            if (rc != CKR_OK) {
-                TRACE_ERROR("Failed to release Process Lock.\n");
-                goto out_nolock;
-            }
-
             init_token_data(tokdata, slot_id);
 
-            rc = XProcLock(tokdata);
-            if (rc != CKR_OK) {
-                TRACE_ERROR("Failed to get Process Lock.\n");
-                goto out_nolock;
-            }
-
             fp = fopen(fname, "r");
             if (!fp) {
                 // were really hosed here since the created
@@ -689,8 +712,7 @@ CK_RV save_token_data_old(STDLL_TokData_t *tokdata, CK_SLOT_ID slot_id)
         goto out_nolock;
     }
 
-    sprintf(fname, "%s/%s", tokdata->data_store, PK_LITE_NV);
-    fp = fopen(fname, "w");
+    fp = open_token_nvdat(fname, sizeof(fname), tokdata, "w");
     if (!fp) {
         TRACE_ERROR("fopen(%s): %s\n", fname, strerror(errno));
         rc = CKR_FUNCTION_FAILED;
@@ -819,8 +841,12 @@ static CK_RV save_private_token_object_old(STDLL_TokData_t *tokdata, OBJECT *obj
         goto error;
     }
 
-    sprintf(fname, "%s/%s/", tokdata->data_store, PK_LITE_OBJ_DIR);
-    strncat(fname, (char *) obj->name, 8);
+    if (ock_snprintf(fname, PATH_MAX, "%s/%s/%.8s", tokdata->data_store,
+                     PK_LITE_OBJ_DIR, (char *)obj->name) != 0) {
+        TRACE_ERROR("private token object old name buffer overflow\n");
+        rc = CKR_FUNCTION_FAILED;
+        goto error;
+    }
     fp = fopen(fname, "w");
     if (!fp) {
         TRACE_ERROR("fopen(%s): %s\n", fname, strerror(errno));
@@ -879,20 +905,14 @@ CK_RV load_private_token_objects_old(STDLL_TokData_t *tokdata)
     CK_RV rc;
     size_t read_size;
 
-    sprintf(iname, "%s/%s/%s", tokdata->data_store, PK_LITE_OBJ_DIR,
-            PK_LITE_OBJ_IDX);
-
-    fp1 = fopen(iname, "r");
+    fp1 = open_token_object_index(iname, sizeof(iname), tokdata, "r");
     if (!fp1)
         return CKR_OK;          // no token objects
 
     while (fgets(tmp, 50, fp1)) {
         tmp[strlen(tmp) - 1] = 0;
 
-        sprintf(fname, "%s/%s/", tokdata->data_store, PK_LITE_OBJ_DIR);
-        strcat(fname, tmp);
-
-        fp2 = fopen(fname, "r");
+        fp2 = open_token_object_path(fname, sizeof(fname), tokdata, tmp, "r");
         if (!fp2)
             continue;
 
@@ -929,10 +949,8 @@ CK_RV load_private_token_objects_old(STDLL_TokData_t *tokdata)
                        "Cannot read token object %s " "(ignoring it)", fname);
             continue;
         }
-        // Grab object list  mutex
-        MY_LockMutex(&obj_list_mutex);
+
         rc = restore_private_token_object_old(tokdata, buf, size, NULL);
-        MY_UnlockMutex(&obj_list_mutex);
         if (rc != CKR_OK)
             goto error;
 
@@ -971,20 +989,44 @@ CK_RV load_masterkey_so_old(STDLL_TokData_t *tokdata)
     CK_ULONG key_len = 0L;
     CK_ULONG master_key_len = 0L;
     CK_ULONG block_size = 0L;
+    struct stat sb;
 
-    if ((rc = get_encryption_info_for_clear_key(&key_len,
-                                                &block_size)) != CKR_OK)
-        goto done;
-
-    if ((rc = get_encryption_info(&master_key_len, NULL)) != CKR_OK)
+    if ((rc = get_encryption_info(&key_len, &block_size)) != CKR_OK)
         goto done;
 
+    master_key_len = key_len;
     memset(tokdata->master_key, 0x0, master_key_len);
 
     data_len = master_key_len + SHA1_HASH_SIZE;
     clear_len = cipher_len = (data_len + block_size - 1)
         & ~(block_size - 1);
 
+    sprintf(fname, "%s/MK_SO", tokdata->data_store);
+    if (stat(fname, &sb) != 0) {
+        TRACE_ERROR("stat(%s): %s\n", fname, strerror(errno));
+        rc = CKR_FUNCTION_FAILED;
+        goto done;
+    }
+
+    if ((CK_ULONG)sb.st_size > cipher_len &&
+        token_specific.secure_key_token &&
+        strcmp(token_specific.token_subdir, "ccatok") == 0) {
+        /*
+         * The CCA token used to have a secure master key length of 64, although
+         * it uses clear keys for the master key in the meantime. The master key
+         * length  has an influence on the file size of the MK_SO and MK_USER
+         * files when using the old pin encryption format. Use special handling
+         * for such larger MK_SO files, and accept the larger length. Newly
+         * written MK_SO files will use the clear key master key length, but we
+         * need to be able to read larger files for backwards compatibility.
+         */
+        master_key_len = 64;
+
+        data_len = master_key_len + SHA1_HASH_SIZE;
+        clear_len = cipher_len = (data_len + block_size - 1)
+             & ~(block_size - 1);
+    }
+
     key = malloc(key_len);
     cipher = malloc(cipher_len);
     clear = malloc(clear_len);
@@ -995,10 +1037,8 @@ CK_RV load_masterkey_so_old(STDLL_TokData_t *tokdata)
     // this file gets created on C_InitToken so we can assume that it always
     // exists
     //
-    sprintf(fname, "%s/MK_SO", tokdata->data_store);
-    fp = fopen(fname, "r");
+    fp = open_token_data_store_path(fname, sizeof(fname), tokdata, "MK_SO", "r");
     if (!fp) {
-        TRACE_ERROR("fopen(%s): %s\n", fname, strerror(errno));
         rc = CKR_FUNCTION_FAILED;
         goto done;
     }
@@ -1074,20 +1114,44 @@ CK_RV load_masterkey_user_old(STDLL_TokData_t *tokdata)
     CK_ULONG key_len = 0L;
     CK_ULONG master_key_len = 0L;
     CK_ULONG block_size = 0L;
+    struct stat sb;
 
-    if ((rc =
-         get_encryption_info_for_clear_key(&key_len, &block_size)) != CKR_OK)
-        goto done;
-
-    if ((rc = get_encryption_info(&master_key_len, NULL)) != CKR_OK)
+    if ((rc = get_encryption_info(&key_len, &block_size)) != CKR_OK)
         goto done;
 
+    master_key_len = key_len;
     memset(tokdata->master_key, 0x0, master_key_len);
 
     data_len = master_key_len + SHA1_HASH_SIZE;
     clear_len = cipher_len = (data_len + block_size - 1)
         & ~(block_size - 1);
 
+    sprintf(fname, "%s/MK_USER", tokdata->data_store);
+    if (stat(fname, &sb) != 0) {
+        TRACE_ERROR("stat(%s): %s\n", fname, strerror(errno));
+        rc = CKR_FUNCTION_FAILED;
+        goto done;
+    }
+
+    if ((CK_ULONG)sb.st_size > cipher_len &&
+        token_specific.secure_key_token &&
+        strcmp(token_specific.token_subdir, "ccatok") == 0) {
+        /*
+         * The CCA token used to have a secure master key length of 64, although
+         * it uses clear keys for the master key in the meantime. The master key
+         * length  has an influence on the file size of the MK_SO and MK_USER
+         * files when using the old pin encryption format. Use special handling
+         * for such larger MK_USER files, and accept the larger length. Newly
+         * written MK_USER files will use the clear key master key length, but
+         * we need to be able to read larger files for backwards compatibility.
+         */
+        master_key_len = 64;
+
+        data_len = master_key_len + SHA1_HASH_SIZE;
+        clear_len = cipher_len = (data_len + block_size - 1)
+             & ~(block_size - 1);
+    }
+
     key = malloc(key_len);
     cipher = malloc(cipher_len);
     clear = malloc(clear_len);
@@ -1098,10 +1162,8 @@ CK_RV load_masterkey_user_old(STDLL_TokData_t *tokdata)
     // this file gets created on C_InitToken so we can assume that it always
     // exists
     //
-    sprintf(fname, "%s/MK_USER", tokdata->data_store);
-    fp = fopen(fname, "r");
+    fp = open_token_data_store_path(fname, sizeof(fname), tokdata, "MK_USER", "r");
     if (!fp) {
-        TRACE_ERROR("fopen(%s): %s\n", fname, strerror(errno));
         rc = CKR_FUNCTION_FAILED;
         goto done;
     }
@@ -1172,7 +1234,6 @@ CK_RV save_masterkey_so_old(STDLL_TokData_t *tokdata)
     CK_ULONG cipher_len = 0L;
     CK_BYTE *key = NULL;
     CK_ULONG key_len = 0L;
-    CK_ULONG master_key_len = 0L;
     CK_ULONG block_size = 0L;
     CK_ULONG data_len = 0L;
     char fname[PATH_MAX];
@@ -1182,14 +1243,10 @@ CK_RV save_masterkey_so_old(STDLL_TokData_t *tokdata)
     if (!token_specific.data_store.use_master_key)
         return CKR_OK;
 
-    if ((rc = get_encryption_info_for_clear_key(&key_len,
-                                                &block_size)) != CKR_OK)
+    if ((rc = get_encryption_info(&key_len, &block_size)) != CKR_OK)
         goto done;
 
-    if ((rc = get_encryption_info(&master_key_len, NULL)) != CKR_OK)
-        goto done;
-
-    data_len = master_key_len + SHA1_HASH_SIZE;
+    data_len = key_len + SHA1_HASH_SIZE;
     cipher_len = clear_len = block_size * (data_len / block_size + 1);
 
     key = malloc(key_len);
@@ -1201,9 +1258,9 @@ CK_RV save_masterkey_so_old(STDLL_TokData_t *tokdata)
         goto done;
     }
     // Copy data to buffer (key+hash)
-    memcpy(clear, tokdata->master_key, master_key_len);
+    memcpy(clear, tokdata->master_key, key_len);
     if ((rc = compute_sha1(tokdata, tokdata->master_key,
-                           master_key_len, clear + master_key_len)) != CKR_OK)
+                           key_len, clear + key_len)) != CKR_OK)
         goto done;
     add_pkcs_padding(clear + data_len, block_size, data_len, clear_len);
 
@@ -1222,10 +1279,8 @@ CK_RV save_masterkey_so_old(STDLL_TokData_t *tokdata)
     //
     // probably ought to ensure the permissions are correct
     //
-    sprintf(fname, "%s/MK_SO", tokdata->data_store);
-    fp = fopen(fname, "w");
+    fp = open_token_data_store_path(fname, sizeof(fname), tokdata, "MK_SO", "w");
     if (!fp) {
-        TRACE_ERROR("fopen(%s): %s\n", fname, strerror(errno));
         rc = CKR_FUNCTION_FAILED;
         goto done;
     }
@@ -1264,20 +1319,15 @@ CK_RV save_masterkey_user_old(STDLL_TokData_t *tokdata)
     CK_ULONG cipher_len = 0L;
     CK_BYTE *key = NULL;
     CK_ULONG key_len = 0L;
-    CK_ULONG master_key_len = 0L;
     CK_ULONG block_size = 0L;
     CK_ULONG data_len = 0L;
     char fname[PATH_MAX];
     CK_RV rc;
 
-    if ((rc = get_encryption_info_for_clear_key(&key_len,
-                                                &block_size)) != CKR_OK)
+    if ((rc = get_encryption_info(&key_len, &block_size)) != CKR_OK)
         goto done;
 
-    if ((rc = get_encryption_info(&master_key_len, NULL)) != CKR_OK)
-        goto done;
-
-    data_len = master_key_len + SHA1_HASH_SIZE;
+    data_len = key_len + SHA1_HASH_SIZE;
     cipher_len = clear_len = block_size * (data_len / block_size + 1);
 
     key = malloc(key_len);
@@ -1289,9 +1339,9 @@ CK_RV save_masterkey_user_old(STDLL_TokData_t *tokdata)
         goto done;
     }
     // Copy data to buffer (key+hash)
-    memcpy(clear, tokdata->master_key, master_key_len);
+    memcpy(clear, tokdata->master_key, key_len);
     if ((rc = compute_sha1(tokdata, tokdata->master_key,
-                           master_key_len, clear + master_key_len)) != CKR_OK)
+                           key_len, clear + key_len)) != CKR_OK)
         goto done;
     add_pkcs_padding(clear + data_len, block_size, data_len, clear_len);
 
@@ -1310,10 +1360,8 @@ CK_RV save_masterkey_user_old(STDLL_TokData_t *tokdata)
     //
     // probably ought to ensure the permissions are correct
     //
-    sprintf(fname, "%s/MK_USER", tokdata->data_store);
-    fp = fopen(fname, "w");
+    fp = open_token_data_store_path(fname, sizeof(fname), tokdata, "MK_USER", "w");
     if (!fp) {
-        TRACE_ERROR("fopen(%s): %s\n", fname, strerror(errno));
         rc = CKR_FUNCTION_FAILED;
         goto done;
     }
@@ -1345,20 +1393,21 @@ CK_RV generate_master_key_old(STDLL_TokData_t *tokdata, CK_BYTE *key)
 {
     CK_RV rc = CKR_OK;
     CK_ULONG key_len = 0L;
-    CK_ULONG master_key_len = 0L;
+    CK_ULONG master_key_len;
+    CK_BYTE *master_key = NULL;
+    CK_BBOOL is_opaque = FALSE;
 
     /* Skip it if master key is not needed. */
     if (!token_specific.data_store.use_master_key)
         return CKR_OK;
 
-    if ((rc = get_encryption_info_for_clear_key(&key_len, NULL)) != CKR_OK ||
-        (rc = get_encryption_info(&master_key_len, NULL)) != CKR_OK)
-        return ERR_FUNCTION_FAILED;
+    if ((rc = get_encryption_info(&key_len, NULL)) != CKR_OK)
+        return CKR_FUNCTION_FAILED;
 
     /* For secure key tokens, object encrypt/decrypt uses
      * software(openssl), not token. So generate masterkey via RNG.
      */
-    if (is_secure_key_token())
+    if (token_specific.secure_key_token)
         return rng_generate(tokdata, key, key_len);
 
     /* For clear key tokens, let token generate masterkey
@@ -1366,14 +1415,32 @@ CK_RV generate_master_key_old(STDLL_TokData_t *tokdata, CK_BYTE *key)
      */
     switch (token_specific.data_store.encryption_algorithm) {
     case CKM_DES3_CBC:
-        return token_specific.t_des_key_gen(tokdata, key,
-                                            master_key_len, key_len);
+        rc = token_specific.t_des_key_gen(tokdata, &master_key,
+                                          &master_key_len, key_len,
+                                          &is_opaque);
+        break;
     case CKM_AES_CBC:
-        return token_specific.t_aes_key_gen(tokdata, key,
-                                            master_key_len, key_len);
+        rc = token_specific.t_aes_key_gen(tokdata, &master_key,
+                                          &master_key_len, key_len,
+                                          &is_opaque);
+        break;
+    default:
+        return CKR_MECHANISM_INVALID;
     }
 
-    return ERR_MECHANISM_INVALID;
+    if (rc != CKR_OK)
+        return rc;
+
+    if (master_key_len != key_len) {
+        TRACE_ERROR("Invalid master key size: %lu\n", master_key_len);
+        free(master_key);
+        return CKR_FUNCTION_FAILED;
+    }
+
+    memcpy(key, master_key, master_key_len);
+    free(master_key);
+
+    return CKR_OK;
 }
 
 CK_RV restore_private_token_object_old(STDLL_TokData_t *tokdata, CK_BYTE *data,
@@ -1496,12 +1563,12 @@ CK_RV reload_token_object_old(STDLL_TokData_t *tokdata, OBJECT *obj)
     CK_RV rc;
     size_t read_size;
 
-    memset(fname, 0x0, sizeof(fname));
-
-    sprintf(fname, "%s/%s/", tokdata->data_store, PK_LITE_OBJ_DIR);
-
-    strncat(fname, (char *) obj->name, 8);
-
+    if (ock_snprintf(fname, PATH_MAX, "%s/%s/%.8s", tokdata->data_store,
+                     PK_LITE_OBJ_DIR, (char *)obj->name) != 0) {
+        TRACE_ERROR("token object file name buffer overflow\n");
+        rc = CKR_FUNCTION_FAILED;
+        goto done;
+    }
     fp = fopen(fname, "r");
     if (!fp) {
         TRACE_ERROR("fopen(%s): %s\n", fname, strerror(errno));
@@ -1571,13 +1638,18 @@ CK_RV save_public_token_object_old(STDLL_TokData_t *tokdata, OBJECT * obj)
     CK_BBOOL flag = FALSE;
     CK_RV rc;
     CK_ULONG_32 total_len;
+
     rc = object_flatten(obj, &clear, &clear_len);
     if (rc != CKR_OK) {
         goto error;
     }
 
-    sprintf(fname, "%s/%s/", tokdata->data_store, PK_LITE_OBJ_DIR);
-    strncat(fname, (char *) obj->name, 8);
+    if (ock_snprintf(fname, PATH_MAX, "%s/%s/%.8s", tokdata->data_store,
+                     PK_LITE_OBJ_DIR, (char *)obj->name) != 0) {
+        TRACE_ERROR("public token object file name buffer overflow\n");
+        rc = CKR_FUNCTION_FAILED;
+        goto error;
+    }
     fp = fopen(fname, "w");
     if (!fp) {
         TRACE_ERROR("fopen(%s): %s\n", fname, strerror(errno));
@@ -1621,20 +1693,14 @@ CK_RV load_public_token_objects_old(STDLL_TokData_t *tokdata)
     CK_ULONG_32 size;
     size_t read_size;
 
-    sprintf(iname, "%s/%s/%s", tokdata->data_store, PK_LITE_OBJ_DIR,
-            PK_LITE_OBJ_IDX);
-
-    fp1 = fopen(iname, "r");
+    fp1 = open_token_object_index(iname, sizeof(iname), tokdata, "r");
     if (!fp1)
         return CKR_OK;          // no token objects
 
     while (fgets(tmp, 50, fp1)) {
         tmp[strlen(tmp) - 1] = 0;
 
-        sprintf(fname, "%s/%s/", tokdata->data_store, PK_LITE_OBJ_DIR);
-        strcat(fname, tmp);
-
-        fp2 = fopen(fname, "r");
+        fp2 = open_token_object_path(fname, sizeof(fname), tokdata, tmp, "r");
         if (!fp2)
             continue;
 
@@ -1718,7 +1784,7 @@ static CK_RV aes_256_gcm_seal(unsigned char *out,
         || EVP_CipherInit_ex(ctx, NULL, NULL, key, iv, 1) != 1
         || EVP_CipherUpdate(ctx, NULL, &outlen, aad, aadlen) != 1
         || EVP_CipherUpdate(ctx, out, &outlen, in, inlen) != 1
-        || EVP_CipherFinal_ex(ctx, out, &outlen) != 1
+        || EVP_CipherFinal_ex(ctx, out + outlen, &outlen) != 1
         || EVP_CIPHER_CTX_ctrl(ctx, EVP_CTRL_GCM_GET_TAG, 16, tag) != 1) {
         TRACE_ERROR("%s\n", ock_err(ERR_GENERAL_ERROR));
         rc = ERR_GENERAL_ERROR;
@@ -1758,7 +1824,7 @@ static CK_RV aes_256_gcm_unseal(unsigned char *out,
         || EVP_CipherInit_ex(ctx, NULL, NULL, key, iv, 0) != 1
         || EVP_CipherUpdate(ctx, NULL, &outlen, aad, aadlen) != 1
         || EVP_CipherUpdate(ctx, out, &outlen, in, inlen) != 1
-        || EVP_CipherFinal_ex(ctx, out, &outlen) != 1) {
+        || EVP_CipherFinal_ex(ctx, out + outlen, &outlen) != 1) {
         TRACE_ERROR("%s\n", ock_err(ERR_GENERAL_ERROR));
         rc = ERR_GENERAL_ERROR;
         goto done;
@@ -1776,6 +1842,7 @@ static CK_RV aes_256_wrap(unsigned char out[40],
 {
     CK_RV rc;
     int outlen;
+    unsigned char buffer[40 + EVP_MAX_BLOCK_LENGTH];
 
     EVP_CIPHER_CTX *ctx = NULL;
 
@@ -1789,13 +1856,14 @@ static CK_RV aes_256_wrap(unsigned char out[40],
     EVP_CIPHER_CTX_set_flags(ctx, EVP_CIPHER_CTX_FLAG_WRAP_ALLOW);
 
     if (EVP_CipherInit_ex(ctx, EVP_aes_256_wrap(), NULL, kek, NULL, 1) != 1
-        || EVP_CipherUpdate(ctx, out, &outlen, in, 32) != 1
-        || EVP_CipherFinal_ex(ctx, out, &outlen) != 1) {
+        || EVP_CipherUpdate(ctx, buffer, &outlen, in, 32) != 1
+        || EVP_CipherFinal_ex(ctx, buffer + outlen, &outlen) != 1) {
         TRACE_ERROR("%s\n", ock_err(ERR_GENERAL_ERROR));
         rc = ERR_GENERAL_ERROR;
         goto done;
     }
 
+    memcpy(out, buffer, 40);
     rc = CKR_OK;
 done:
     EVP_CIPHER_CTX_free(ctx);
@@ -1808,6 +1876,7 @@ static CK_RV aes_256_unwrap(unsigned char key[32],
 {
     CK_RV rc;
     int outlen;
+    unsigned char buffer[32 + EVP_MAX_BLOCK_LENGTH];
 
     EVP_CIPHER_CTX *ctx = NULL;
 
@@ -1821,13 +1890,14 @@ static CK_RV aes_256_unwrap(unsigned char key[32],
     EVP_CIPHER_CTX_set_flags(ctx, EVP_CIPHER_CTX_FLAG_WRAP_ALLOW);
 
     if (EVP_CipherInit_ex(ctx, EVP_aes_256_wrap(), NULL, kek, NULL, 0) != 1
-        || EVP_CipherUpdate(ctx, key, &outlen, in, 40) != 1
-        || EVP_CipherFinal_ex(ctx, key, &outlen) != 1) {
+        || EVP_CipherUpdate(ctx, buffer, &outlen, in, 40) != 1
+        || EVP_CipherFinal_ex(ctx, buffer + outlen, &outlen) != 1) {
         TRACE_ERROR("%s\n", ock_err(ERR_GENERAL_ERROR));
         rc = ERR_GENERAL_ERROR;
         goto done;
     }
 
+    memcpy(key, buffer, 32);
     rc = CKR_OK;
 done:
     EVP_CIPHER_CTX_free(ctx);
@@ -1871,10 +1941,9 @@ CK_RV save_masterkey_so(STDLL_TokData_t *tokdata)
     //
     // probably ought to ensure the permissions are correct
     //
-    sprintf(fname, "%s/MK_SO", tokdata->data_store);
-    fp = fopen(fname, "w");
+    fp = open_token_data_store_path(fname, sizeof(fname), tokdata, "MK_SO",
+                                    "w");
     if (!fp) {
-        TRACE_ERROR("fopen(%s): %s\n", fname, strerror(errno));
         rc = CKR_FUNCTION_FAILED;
         goto done;
     }
@@ -1909,10 +1978,9 @@ CK_RV load_masterkey_so(STDLL_TokData_t *tokdata)
     // this file gets created on C_InitToken so we can assume that it always
     // exists
     //
-    sprintf(fname, "%s/MK_SO", tokdata->data_store);
-    fp = fopen(fname, "r");
+    fp = open_token_data_store_path(fname, sizeof(fname), tokdata, "MK_SO",
+                                    "r");
     if (!fp) {
-        TRACE_ERROR("fopen(%s): %s\n", fname, strerror(errno));
         rc = CKR_FUNCTION_FAILED;
         goto done;
     }
@@ -1961,10 +2029,9 @@ CK_RV save_masterkey_user(STDLL_TokData_t *tokdata)
     //
     // probably ought to ensure the permissions are correct
     //
-    sprintf(fname, "%s/MK_USER", tokdata->data_store);
-    fp = fopen(fname, "w");
+    fp = open_token_data_store_path(fname, sizeof(fname), tokdata, "MK_USER",
+                                    "w");
     if (!fp) {
-        TRACE_ERROR("fopen(%s): %s\n", fname, strerror(errno));
         rc = CKR_FUNCTION_FAILED;
         goto done;
     }
@@ -2000,10 +2067,9 @@ CK_RV load_masterkey_user(STDLL_TokData_t *tokdata)
     // this file gets created on C_InitToken so we can assume that it always
     // exists
     //
-    sprintf(fname, "%s/MK_USER", tokdata->data_store);
-    fp = fopen(fname, "r");
+    fp = open_token_data_store_path(fname, sizeof(fname), tokdata, "MK_USER",
+                                    "r");
     if (!fp) {
-        TRACE_ERROR("fopen(%s): %s\n", fname, strerror(errno));
         rc = CKR_FUNCTION_FAILED;
         goto done;
     }
@@ -2045,8 +2111,7 @@ CK_RV save_token_data(STDLL_TokData_t *tokdata, CK_SLOT_ID slot_id)
         goto out_nolock;
     }
 
-    sprintf(fname, "%s/%s", tokdata->data_store, PK_LITE_NV);
-    fp = fopen(fname, "w");
+    fp = open_token_nvdat(fname, sizeof(fname), tokdata, "w");
     if (!fp) {
         TRACE_ERROR("fopen(%s): %s\n", fname, strerror(errno));
         rc = CKR_FUNCTION_FAILED;
@@ -2133,28 +2198,12 @@ CK_RV load_token_data(STDLL_TokData_t *tokdata, CK_SLOT_ID slot_id)
         goto out_nolock;
     }
 
-    sprintf(fname, "%s/%s", tokdata->data_store, PK_LITE_NV);
-    fp = fopen(fname, "r");
+    fp = open_token_nvdat(fname, sizeof(fname), tokdata, "r");
     if (!fp) {
         /* Better error checking added */
         if (errno == ENOENT) {
-            /* init_token_data may call save_token_data, which
-             * grabs the lock, so we must release it around this
-             * call */
-            rc = XProcUnLock(tokdata);
-            if (rc != CKR_OK) {
-                TRACE_ERROR("Failed to release Process Lock.\n");
-                goto out_nolock;
-            }
-
             init_token_data(tokdata, slot_id);
 
-            rc = XProcLock(tokdata);
-            if (rc != CKR_OK) {
-                TRACE_ERROR("Failed to get Process Lock.\n");
-                goto out_nolock;
-            }
-
             fp = fopen(fname, "r");
             if (!fp) {
                 // were really hosed here since the created
@@ -2274,8 +2323,8 @@ static inline int inc32(unsigned char ctr[4])
  * u8 tag[16]                    | 16-byte footer
  * ----------------           <--+
  */
-#define HEADER_LEN	64
-#define FOOTER_LEN	16
+#define HEADER_LEN  64
+#define FOOTER_LEN  16
 
 //
 // Note: The token lock (XProcLock) must be held when calling this function.
@@ -2443,20 +2492,14 @@ CK_RV load_private_token_objects(STDLL_TokData_t *tokdata)
     if (tokdata->version < TOK_NEW_DATA_STORE)
         return load_private_token_objects_old(tokdata);
 
-    sprintf(iname, "%s/%s/%s", tokdata->data_store, PK_LITE_OBJ_DIR,
-            PK_LITE_OBJ_IDX);
-
-    fp1 = fopen(iname, "r");
+    fp1 = open_token_object_index(iname, sizeof(iname), tokdata, "r");
     if (!fp1)
         return CKR_OK;          // no token objects
 
     while (fgets(tmp, 50, fp1)) {
         tmp[strlen(tmp) - 1] = 0;
 
-        sprintf(fname, "%s/%s/", tokdata->data_store, PK_LITE_OBJ_DIR);
-        strcat(fname, tmp);
-
-        fp2 = fopen(fname, "r");
+        fp2 = open_token_object_path(fname, sizeof(fname), tokdata, tmp,"r");
         if (!fp2)
             continue;
 
@@ -2497,11 +2540,9 @@ CK_RV load_private_token_objects(STDLL_TokData_t *tokdata)
             continue;
         }
 
-        MY_LockMutex(&obj_list_mutex);
         rc = restore_private_token_object(tokdata, header,
                                           buf, size,
                                           footer, NULL);
-        MY_UnlockMutex(&obj_list_mutex);
         if (rc != CKR_OK)
             goto error;
 
@@ -2587,6 +2628,7 @@ CK_RV reload_token_object(STDLL_TokData_t *tokdata, OBJECT *obj)
     CK_ULONG size_64;
     CK_RV rc;
     uint32_t len;
+    uint32_t ver;
 
     if (tokdata->version < TOK_NEW_DATA_STORE)
         return reload_token_object_old(tokdata, obj);
@@ -2610,9 +2652,18 @@ CK_RV reload_token_object(STDLL_TokData_t *tokdata, OBJECT *obj)
         goto done;
     }
 
+    memcpy(&ver, header, 4);
     memcpy(&priv, header + 4, 1);
     memcpy(&len, header + 60, 4);
-    size = be32toh(len);
+
+    /*
+     * In OCK 3.12 - 3.14 the version and size was not stored in BE. So if
+     * version field is in platform endianness, keep size as is also.
+     */
+    if (ver == TOK_NEW_DATA_STORE)
+        size = len;
+    else
+        size = be32toh(len);
 
     buf = (CK_BYTE *) malloc(size);
     if (buf == NULL) {
@@ -2664,7 +2715,7 @@ done:
  * u8  object[object_len]        | body
  * ----------------           <--+
  */
-#define PUB_HEADER_LEN	16
+#define PUB_HEADER_LEN  16
 
 //
 // Note: The token lock (XProcLock) must be held when calling this function.
@@ -2677,8 +2728,9 @@ CK_RV save_public_token_object(STDLL_TokData_t *tokdata, OBJECT *obj)
     CK_ULONG clear_len;
     CK_BBOOL flag = FALSE;
     CK_RV rc;
-    CK_ULONG_32 len;
+    CK_ULONG_32 len, be_len;
     unsigned char reserved[7] = {0};
+    uint32_t tmp;
 
     if (tokdata->version < TOK_NEW_DATA_STORE)
         return save_public_token_object_old(tokdata, obj);
@@ -2699,11 +2751,14 @@ CK_RV save_public_token_object(STDLL_TokData_t *tokdata, OBJECT *obj)
         goto done;
     }
 
+    tmp = htobe32(tokdata->version);
+    be_len = htobe32(len);
+
     set_perm(fileno(fp));
-    if (fwrite(&tokdata->version, 4, 1, fp) != 1
+    if (fwrite(&tmp, 4, 1, fp) != 1
         || fwrite(&flag, 1, 1, fp) != 1
         || fwrite(reserved, 7, 1, fp) != 1
-        || fwrite(&len, 4, 1, fp) != 1
+        || fwrite(&be_len, 4, 1, fp) != 1
         || fwrite(clear, len, 1, fp) != 1) {
         rc = CKR_FUNCTION_FAILED;
         goto done;
@@ -2734,14 +2789,12 @@ CK_RV load_public_token_objects(STDLL_TokData_t *tokdata)
     CK_BBOOL priv;
     CK_ULONG_32 size;
     unsigned char header[PUB_HEADER_LEN];
+    uint32_t ver;
 
     if (tokdata->version < TOK_NEW_DATA_STORE)
         return load_public_token_objects_old(tokdata);
 
-    sprintf(iname, "%s/%s/%s", tokdata->data_store, PK_LITE_OBJ_DIR,
-            PK_LITE_OBJ_IDX);
-
-    fp1 = fopen(iname, "r");
+    fp1 = open_token_object_index(iname, sizeof(iname), tokdata, "r");
     if (!fp1)
         return CKR_OK;          // no token objects
 
@@ -2761,9 +2814,16 @@ CK_RV load_public_token_objects(STDLL_TokData_t *tokdata)
             continue;
         }
 
+        memcpy(&ver, header, 4);
         memcpy(&priv, header + 4, 1);
         memcpy(&size, header + 12, 4);
-        size = be32toh(size);
+
+        /*
+         * In OCK 3.12 - 3.14 the version and size was not stored in BE. So if
+         * version field is in platform endianness, keep size as is also
+         */
+        if (ver != TOK_NEW_DATA_STORE)
+            size = be32toh(size);
 
         if (priv == TRUE) {
             fclose(fp2);
